#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "gfx.h"
#include "pngio.h"

static int dW,dH;
static unsigned int *dD;

static int iW,iH;
static unsigned int *iD;

static int clip_black(unsigned int *D, unsigned int W, unsigned int H,
                       unsigned int **Rd, unsigned int *Rw, unsigned int *Rh)
{
  int top,lef,rig,bot;
  int x,y,nW,nH;
  unsigned int *tD;
  unsigned int p;
  p= 0; // D[0];

  for(top=0;top<H;top++)     for(x=0;x<W;x++) if (D[x+top*W]!=p) goto L1;
L1:
  for(bot=0;bot+top<H;bot++) for(x=0;x<W;x++) if (D[x+(H-1-bot)*W]!=p) goto L2;
L2:
  for(lef=0;lef<W;lef++)     for(y=0;y<H;y++) if (D[lef+y*W]!=p) goto L3;
L3:
  for(rig=0;rig+lef<W;rig++) for(y=0;y<H;y++) if (D[(W-1-rig)+y*W]!=p) goto L4;
L4:

  nW= W - (lef+rig);
  nH= H - (top+bot);

  if (nH<=0 || nW<=0) return 0;

  tD= malloc(sizeof(unsigned int)*nW*nH);

  for(y=0;y<nH;y++)
    for(x=0;x<nW;x++)
      tD[x+y*nW]= D[(x+lef)+(y+top)*W];

  *Rd= tD;
  *Rw= nW;
  *Rh= nH;
  return 1;
}

unsigned int combine(unsigned int V,unsigned int U,double F)
{
  int R,G,B;

  B= (V&0xff)+ (U&0xff)*F; V>>= 8; U>>= 8;
  G= (V&0xff)+ (U&0xff)*F; V>>= 8; U>>= 8;
  R= (V&0xff)+ (U&0xff)*F; 

  B= B<0 ? 0 : (B>255 ? 255 : B);
  G= G<0 ? 0 : (G>255 ? 255 : G);
  R= R<0 ? 0 : (R>255 ? 255 : R);

  return (R<<16) | (G<<8) | B;
}

static void rotate_pix(unsigned int *D, unsigned int W, unsigned int H,
                       int angle_degrees,
                       unsigned int **Rd, unsigned int *Rw, unsigned int *Rh)
{
  double cosA, sinA;
  unsigned int tW, tH,*tD;
  int X,Y;
  double G,C,S,V;
  int pX,pY;
  double Xp,Yp;
  double F0,F1,F2,F3;
  int p0,p1,p2,p3;
  int cB, cG, cR;
  double Xt;
  int oob= 0;
  int cx,cy,cw,ch, dxdx, dxdy, dydy, dydx, xo,yo;

  while(angle_degrees<0) angle_degrees+= 360;
  angle_degrees %= 360;
  switch(angle_degrees/90)
  {
case 0: dxdy=  0; dxdx=  1; xo= 0;   dydx=  0; dydy=  1; yo= 0;  
                                                            cw=W; ch=H; break;
case 1: dxdy= -1; dxdx=  0; xo= H-1; dydx=  1; dydy=  0; yo= 0;
                                                            cw=H; ch=W; break;
case 2: dxdy=  0; dxdx= -1; xo= W-1; dydx=  0; dydy= -1; yo= H-1; 
                                                            cw=W; ch=H; break;
case 3: dxdy=  1; dxdx=  0; xo= 0;   dydx= -1; dydy=  0; yo= W-1;
                                                            cw=H; ch=W; break;
  }

  angle_degrees%= 90;

  cosA= cos(angle_degrees/180.0*M_PI);
  sinA= sin(angle_degrees/180.0*M_PI);

  printf("cos= %f sin=%f\n",cosA,sinA);


  tW= ceil(cw*cosA + ch*sinA)+4;
  tH= ceil(ch*cosA + cw*sinA)+4;
  tD= calloc(tW*tH,sizeof(unsigned int));

  printf("w= %d h= %d . cw= %d ch= %d . tw= %d th= %d\n", W,H,cw,ch,tW,tH);

  for(Y=0;Y<H;Y++)
   for(X=0;X<W;X++)
   {
     cx= X*dxdx + Y*dxdy + xo;
     cy= X*dydx + Y*dydy + yo;

     Xp= cx*cosA-cy*sinA + ch*sinA;
     Yp= cx*sinA+cy*cosA;

     G= ((int)Xp+1)-Xp;
     C= ((int)Yp+1)-Yp;

     pX= Xp+1;
     pY= Yp+1;

     if (pX>=0 && pY>=0 && pX<tW && pY<tH) 
         tD[pX+pY*tW]= combine(tD[pX+pY*tW], D[X+Y*W], G*C);
     else oob++;

     pX++;
     if (pX>=0 && pY>=0 && pX<tW && pY<tH) 
         tD[pX+pY*tW]= combine(tD[pX+pY*tW], D[X+Y*W], (1-G)*C);
     else oob++;
 
     pY++;
     if (pX>=0 && pY>=0 && pX<tW && pY<tH) 
         tD[pX+pY*tW]= combine(tD[pX+pY*tW], D[X+Y*W], (1-G)*(1-C));
     else oob++;

     pX--;
     if (pX>=0 && pY>=0 && pX<tW && pY<tH) 
         tD[pX+pY*tW]= combine(tD[pX+pY*tW], D[X+Y*W], G*(1-C));
     else oob++;
   }

   *Rd= tD;
   *Rw= tW;
   *Rh= tH;

   printf("oob= %d\n",oob);
}

void put_image(int w,int h,unsigned int *d)
{
  int x,y;
  int i;
  for(i=0;i<dW*dH;i++) dD[i]= 0xFFffFF;

  if (iD) free(iD);

  iD= d; iW= w; iH= h;

  if (iD)
  {
    for(y=0;y<iH && y<dH;y++)
    for(x=0;x<iW && x<dW;x++)
     dD[x+y*dW]= iD[x+y*iW];
  }
  gfx_pixgrid_putsimple(Wcanvas,dD);
}

void rotim()
{
  char *ate;
  unsigned int rW, rH, *rD;
  unsigned int cW, cH, *cD;
  
  int angle;
  ate= gfx_get_text(Wangle);
  if (!ate || !ate[0]) ate= "5";
  angle= atoi(ate);
  if (!iW || !iH || !iD) return ;

  rotate_image_double(iD, iW, iH, angle, &rD, &rW, &rH);
  if (clip_black(rD, rW, rH, &cD, &cW, &cH))
  {
    free(rD);
    rD= cD; rW= cW; rH= cH;
  }
  put_image(rW,rH,rD);
}

void load_image(char *fn)
{ 
  unsigned int w,h,*d;
  char *e;
  d= NULL;
  w= h= 0;
  e= read_png_image(fn,&w,&h,&d);
  if (!e)
  {
    if (w>dW || h>dH)
    {
      unsigned int rw,rh,*rd;
      resize_image_argb(d,w,h, dW, dH, &rd, &rw, &rh);
      free(d); d= rd; w= rw; h= rh; 
    }
  }
  else printf("can't load %s: %s\n",fn, e);
  put_image(w,h,d);
}



int main(int argc,char **argv)
{
  gfx_event_t event;
  unsigned int w,h;
  int i;
  gfx_init(NULL, 900,600,"Image Rotate");
  gfx_SCREEN_0();
  gfx_pixgrid_attribs(Wcanvas, NULL,&w,&h);
  dD= malloc(sizeof(unsigned int)*w*h);
  for(i=0;i<w*h;i++) dD[i]= 0xFFffff;
  dW= w; dH= h;

  load_image(argv[1]);

  while(1)
  if (!gfx_event(NULL,&event)) continue;
  else switch(event.type)
  {
   case GFXEV_QUIT: goto done;
   case Wrot: rotim(); break;
   case Wload: load_image(argv[1]); break;
  }
done:
  gfx_close();
  return 0;
}
