#include <math.h>
#include <stdlib.h>
#include "rotate_image.h"

static int tables_ok;

static double s2l[256];
static unsigned char l2s[1<<14];
#define TABSIZ sizeof(l2s)

static void init_tables()
{
  int i;
  if (tables_ok) return ;
  tables_ok= 1;
  for(i=0;i<256;i++)
    s2l[i]= pow( (double) i / 255, 2.2 );
  for(i=0;i<TABSIZ;i++)
  {
    double V;
    V= (double) i / (TABSIZ-1);
    l2s[i]= pow(V,1/2.2)*255;
  }
}

static double rgb2linear(unsigned char K)
{
  return s2l[K];
}

static unsigned int linear2rgb(double K)
{
  K= K < 0 ? 0 : (K > 1 ? 1 : K);
  return l2s[ (int) (K*(TABSIZ-1)) ];
}


#if 0

static unsigned int combine(unsigned int V,unsigned int U,unsigned int F)
{
  unsigned int R,G,B;

  B= (((V&0xff)<<16)+ (U&0xff)*F)>>16; V>>= 8; U>>= 8;  
  G= (((V&0xff)<<16)+ (U&0xff)*F)>>16; V>>= 8; U>>= 8; 
  R= (((V&0xff)<<16)+ (U&0xff)*F)>>16;                

  B= B<0 ? 0 : (B>255 ? 255 : B);
  G= G<0 ? 0 : (G>255 ? 255 : G);
  R= R<0 ? 0 : (R>255 ? 255 : R);

  return (R<<16) | (G<<8) | B;
}


void rotate_pixd(unsigned int *D, unsigned int W, unsigned int H,
                       int angle_degrees, 
                       unsigned int **Rd, unsigned int *Rw, unsigned int *Rh)
{
  unsigned int cosA, sinA;
  double cosA_d, sinA_d;
  unsigned int tW, tH,*tD;
  int X,Y;
  int pX,pY;
  int cx,cy,cw,ch, dxdx, dxdy, dydy, dydx, xo,yo;
  unsigned int Xp,Yp,G,C;

  while(angle_degrees<0) angle_degrees+= 360;
  angle_degrees %= 360;
  switch(angle_degrees/90)
  {
case 0: dxdy= 0;dxdx= 1;xo=0;  dydx= 0;dydy= 1;yo=0;   cw=W; ch=H; break;
case 1: dxdy=-1;dxdx= 0;xo=H-1;dydx= 1;dydy= 0;yo=0;   cw=H; ch=W; break;
case 2: dxdy= 0;dxdx=-1;xo=W-1;dydx= 0;dydy=-1;yo=H-1; cw=W; ch=H; break;
case 3: dxdy= 1;dxdx= 0;xo=0;  dydx=-1;dydy= 0;yo=W-1; cw=H; ch=W; break;
  }

  angle_degrees%= 90;

  cosA_d= cos(angle_degrees/180.0*M_PI);
  sinA_d= sin(angle_degrees/180.0*M_PI);

  tW= ceil(cw*cosA_d + ch*sinA_d)+4;
  tH= ceil(ch*cosA_d + cw*sinA_d)+4;
  tD= calloc(tW*tH,sizeof(unsigned int));

  cosA= 256*cosA_d; sinA= 256*sinA_d;

  printf("w= %d h= %d\n",tW,tH);

  for(Y=0;Y<H;Y++)
   for(X=0;X<W;X++)
   {
     cx= X*dxdx + Y*dxdy + xo;
     cy= X*dydx + Y*dydy + yo;

     Xp= cx*cosA-cy*sinA + ch*sinA;
     Yp= cx*sinA+cy*cosA;

     G= 256-(Xp&0xff);
     C= 256-(Yp&0xff);

     pX= ((Xp)>>8)+1;
     pY= ((Yp)>>8)+1;

     tD[pX+pY*tW]= combine(tD[pX+pY*tW], D[X+Y*W], G*C);

     pX++;
     tD[pX+pY*tW]= combine(tD[pX+pY*tW], D[X+Y*W], (256-G)*C);
 
     pY++;
     tD[pX+pY*tW]= combine(tD[pX+pY*tW], D[X+Y*W], (256-G)*(256-C));

     pX--;
     tD[pX+pY*tW]= combine(tD[pX+pY*tW], D[X+Y*W], G*(256-C));
   }

   *Rd= tD;
   *Rw= tW;
   *Rh= tH;
}
#endif

void rotate_image_double(unsigned int *D, unsigned int W, unsigned int H,
                       int angle_degrees, 
                       unsigned int **Rd, unsigned int *Rw, unsigned int *Rh)
{
  double cosA, sinA;
  unsigned int tW, tH,*tD;
  int X,Y;
  int pX,pY;
  int cx,cy,cw,ch, dxdx, dxdy, dydy, dydx, xo,yo;
  double Xp,Yp,G,C;
  double *buffer;
  int shift;
  int i;
  double P;

  init_tables();

  while(angle_degrees<0) angle_degrees+= 360;
  angle_degrees %= 360;
  switch(angle_degrees/90)
  {
case 0: dxdy= 0;dxdx= 1;xo=0;  dydx= 0;dydy= 1;yo=0;   cw=W; ch=H; break;
case 1: dxdy=-1;dxdx= 0;xo=H-1;dydx= 1;dydy= 0;yo=0;   cw=H; ch=W; break;
case 2: dxdy= 0;dxdx=-1;xo=W-1;dydx= 0;dydy=-1;yo=H-1; cw=W; ch=H; break;
case 3: dxdy= 1;dxdx= 0;xo=0;  dydx=-1;dydy= 0;yo=W-1; cw=H; ch=W; break;
  }

  angle_degrees%= 90;

  cosA= cos(angle_degrees/180.0*M_PI);
  sinA= sin(angle_degrees/180.0*M_PI);

  tW= ceil(cw*cosA+ ch*sinA)+4;
  tH= ceil(ch*cosA+ cw*sinA)+4;
  tD= calloc(tW*tH,sizeof(unsigned int));

  buffer= malloc(tW*tH*sizeof(double));

  for(shift=0;shift<24;shift+=8)
  {
    for(i=0;i<tW*tH;i++) buffer[i]= 0.0;

    for(Y=0;Y<H;Y++)
     for(X=0;X<W;X++)
     {
       cx= X*dxdx + Y*dxdy + xo;
       cy= X*dydx + Y*dydy + yo;
  
       Xp= cx*cosA-cy*sinA + ch*sinA;
       Yp= cx*sinA+cy*cosA;

       G= ((int)Xp+1)-Xp;
       C= ((int)Yp+1)-Yp;

       pX= (int) Xp + 1;
       pY= (int) Yp + 1;

       P= rgb2linear( (D[X+Y*W] >> shift)&0xff );

             buffer[pX + pY*tW]+=  P*G*C;
       pX++; buffer[pX + pY*tW]+=  P*(1-G)*C;
       pY++; buffer[pX + pY*tW]+=  P*(1-G)*(1-C);
       pX--; buffer[pX + pY*tW]+=  P*G*(1-C);
     }

     for(i=0;i<tW*tH;i++) tD[i] |= linear2rgb(buffer[i]) << shift;
   }

   free(buffer);

   *Rd= tD;
   *Rw= tW;
   *Rh= tH;
}

